`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Engineer: Scott Gravenhorst
// email: music.maker@gte.net
// 
// Create Date:     11-04-2008 
// Design Name:     Digital Waveguide Polyphonic Synthesizer
// Module Name:     Xarp_55
// Project Name:    Digital Waveguide Polyphonic Synthesizer
// Target Devices:  Spartan-3A DSP 1800
//
// Multiple Karplus-Strong String Model, 56 separate strings, each which can 
// vibrate simultaneously with the others.  Xarp consists of 8 mini-harps.  Each
// mini-harp contains 7 strings.  The mini-harps are processed in parallel 
// allowing a sample rate of 198,143.
//
//////////////////////////////////////////////////////////////////////////////////
//
// CURRENT VERSION (for Spartan-3A DSP)
// String loop Waveguide: 4096 location by 18 bit wide RAMs (block).
// Sample Rate: 198,143 samples per second.  Not variable.
//
//////////////////////////////////////////////////////////////////////////////////
//
//                                PUSH BUTTON CONTROLS:
//              PB[0] reset
//              PB[1] squelch
//              PB[2] squelch
//              PB[3] squelch
//
//              SW[7] octave     0 = high range, 1 = low range
//              SW[6:4] timbre   3 bits determines how wide the excite pulse is.
//                               0 is the widest pulse.
//
///////////////////////////////////////////////////////////////////////////////////
//
// For Spartan-3A DSP
//
// ver_a: WORKING AS OF 11-14-2008
//   harp consists of 8 mini-harps, each of which supports up to 7 strings
//   55 of the 56 available strings are used.  Table begins on a C (about 32 Hz)
//   
// ver_b:
//   - Generate the samples at twice the sample rate then decimate 2:1 should put
//     the instrument one octave higher then ver_a.
//   - Enable sustain.
//   - Modwheel helps to controls filter cutoff
//   - Bug corrected from ver_a: The code to load registers WGxlen with the length
//     of each string was executing one state too early causing wrong timbre, 
//     especially in the first octave.  Now the tone is natural and consistent as 
//     you move up the scale.
//   - 56th string was added, it has 5.051 cents error, but I don't hear it.
//
//////////////////////////////////////////////////////////////////////////////////

module Xarp_56( clk125mhz, 
                led,                                    // led[7:0]
                PB,                                     // PB[0:0]
                SW,                                     // SW[3:0]
                DIGI2_4,                                // DIGI2 Digilent Port J6 (MIDI input)
                DIGI1_1, DIGI1_2, DIGI1_3, DIGI1_4 );   // DIGI1 Digilent Port J7 (I2S DAC)

  input clk125mhz;               // soldered 125 MHz clock
  output [7:0] led;
  input [3:0] PB;                // We use one of the 4 pushbuttons
  input [7:0] SW;                // We use 4 of the DIP switches.

  output DIGI1_1;                 // L18 - I2S DAC connection (J7)
  output DIGI1_2;                 // L17
  output DIGI1_3;                 // E24
  output DIGI1_4;                 // F23
  
  input DIGI2_4;                  // G22 - MIDI connection (J6)
  
/////////////////////////////////////////////////////////////////////////////////////

  wire clk125mhz;          // soldered 125 MHz clock
  wire sysclk;             // 38.043478 MHz - 
  wire [7:0] led;
  wire [7:0] SW;
  wire [3:0] PB;

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
// 38.043478 MHz clock generator
  
  DCM_38_043478 DCMCLK0 (
    .CLKIN_IN( clk125mhz ), 
    .RST_IN( 1'b0 ), 
    .CLKFX_OUT( sysclk ), 
    .CLK0_OUT(), 
    .LOCKED_OUT()
    );
  
  wire DIGI1_1;                 // I2S DAC connection (J7)
  wire DIGI1_2;
  wire DIGI1_3;
  wire DIGI1_4;
  
  wire DIGI2_4;                 // MIDI connection (J6)

  wire [3:0] rstd;              // POR delay
  wire reset;                   // POR/User reset

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////

  wire interrupt;
  wire interrupt_ack;
  wire [9:0] address;             // wires to connect address lines from uC to ROM
  wire [17:0] instruction;        // uC data lines, need connection between uC and ROM
  wire [7:0] out_port;            //
  wire [7:0] in_port;             // 
  wire [7:0] port_id;
  wire proc_reset;                // an output from the mcu ROM, not sure what it's supposed to connect to.
  reg [7:0] in_port_reg;          // hold data for mcu
  assign in_port = in_port_reg;

// MIDI & TTY receivers
  wire [7:0] rx0data;
  wire rx0ready_status;
  wire reset_rx0ready_status;
  
  wire resetsignal;
// mcu
  wire read_strobe;
  wire write_strobe;

//////////////////////////////////////////////////////////////////////////////////////////////////
// Synth signals


// Output values from processing of MIDI data.
  reg [6:0] MOD_WHL = 7'd64;                      // a usable initial value
  
  reg SUSTAIN = 1'b0;
  
  reg [6:0] VEL;               // dist RAM

  reg [2:0] STR  = 0;                            // used as string select within each mini-harp

  reg [2:0] STR0 = 0;                            // used as string select within mini-harp 0
  reg [2:0] STR1 = 0;                            // used as string select within mini-harp 1
  reg [2:0] STR2 = 0;                            // used as string select within mini-harp 2
  reg [2:0] STR3 = 0;                            // used as string select within mini-harp 3
  reg [2:0] STR4 = 0;                            // used as string select within mini-harp 4
  reg [2:0] STR5 = 0;                            // used as string select within mini-harp 5
  reg [2:0] STR6 = 0;                            // used as string select within mini-harp 6
  reg [2:0] STR7 = 0;                            // used as string select within mini-harp 7

//////////////////////////////////////////////////////////////////////////////////////////////////
// POR delay FF chain - taken from Eric Brombaugh's code for the SPI DAC.
  FDCE rst_bit0 (.Q(rstd[0]), .C(sysclk), .CE(1'b1), .CLR(1'b0), .D(1'b1));
  FDCE rst_bit1 (.Q(rstd[1]), .C(sysclk), .CE(1'b1), .CLR(1'b0), .D(rstd[0]));
  FDCE rst_bit2 (.Q(rstd[2]), .C(sysclk), .CE(1'b1), .CLR(1'b0), .D(rstd[1]));
  FDCE rst_bit3 (.Q(rstd[3]), .C(sysclk), .CE(1'b1), .CLR(1'b0), .D(rstd[2]));
  assign reset = ~rstd[3] | PB[0];

//////////////////////////////////////////////////////////////////////////////////////////////////
// instantiate the uC (kcpsm3) with it's ROM

  kcpsm3 MCU0 ( .address(address), .instruction(instruction), .port_id(port_id), 
    .write_strobe(write_strobe), .out_port(out_port), .read_strobe(read_strobe), .in_port(in_port), 
    .interrupt(interrupt), .interrupt_ack(interrupt_ack), 
    .reset(proc_reset|reset), .clk(sysclk) );  

  midictrl PSM_ROM0 ( .address(address), .instruction(instruction), .proc_reset(proc_reset), .clk(sysclk) );
  
  // MIDI receiver.  31,250 Baud
  assign resetsignal = write_strobe && ( port_id == 8'hFF );  // When port_id == FF with write strobe, reset the UARTs

//////////////////////////////////////////////////////////////////////////////////////////////////
// MIDI UART (rx only)
// UART module by Jim Patchell

  wire MIDI_In;
  assign MIDI_In = DIGI2_4;
  
  MIDIuartrx RX_MIDI ( .dout(rx0data), .clk(sysclk), .reset(resetsignal), .rxd(MIDI_In), 
    .frame(), .overrun(), .ready(), .busy(), .CS(), 
    .interrupt(interrupt), .interrupt_ack(interrupt_ack), 
    .rxready_status(rx0ready_status), 
    .reset_rxready_status(reset_rx0ready_status)
    );
  /////// VERY IMPORTANT HARDWARE /////////////////////////////////////////////
  // decode read port 01, send pulse to reset rxready flop
  // This allows the mcu to clear the rxready bit automatically just by reading rxdata.
  assign reset_rx0ready_status = (read_strobe == 1'b1) & (port_id[4:0] == 5'h01);

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// DAC - This DAC version accepts value 'cycles' which determines the amount of time between enables.
// DAC module by Eric Brombaugh

  reg signed [23:0] DACreg = 24'sh000000;    // Data for SPI DAC
  wire signed [23:0] DACbus;

  i2s_out I2SDAC ( .clk( sysclk ), 
                   .reset( reset ), 
                   .l_data( DACreg ), 
                   .r_data( DACreg ),         // mono by sending same data to both ports
                   .sdout( DIGI1_4 ), 
                   .sclk( DIGI1_3 ), 
                   .lrclk( DIGI1_2 ), 
                   .DACena( DACena ) );

  assign DIGI1_1 = sysclk;             // 38.043478 MHz - this is MCLK

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// Pluck logic:
  reg        Set_Pluck = 0;    //
  reg        Clr_Pluck = 0;    // 8 flipflops, one for each mini-harp
  reg  [2:0] Plucked_Str;      // String within a mini-harp which was plucked
  reg  [2:0] Plucked_MH;       // Mini Harp to receive pluck signal
  wire       Pluck;            // Q from Pluck FF

//////////////////////////////////////////////////////////////////////////////////////////////////////////
//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//
//==//==//==//==//==//==//==//==//==//  WAVEGUIDE RAM   //==//==//==//==//==//==//==//==//==//==//==//==//
//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//==//
//////////////////////////////////////////////////////////////////////////////////////////////////////////
/*

lengths per mini harp

Eight 8K x 18 bit RAMs.
55 strings, starting on C (32.703hz) with ~198KHz sample rate

In this setup, the ROM can be address by using an adjusted version of the note number

received by the MIDI controller.  "Adjusted" meaning that the low C on the keyboard

should deliver a value of zero to this hardware.  The MIDI controller is responsible 

for doing this.

0:0	3029  	0:1	1908		0:2	1202		0:3	757		0:4	477		0:5	301		0:6	189
1:0	2859		1:1	1801		1:2	1135		1:3	715		1:4	450		1:5	284		1:6	179
2:0	2699		2:1	1700		2:2	1071		2:3	675		2:4	425		2:5	268		2:6	169
3:0	2547		3:1	1605		3:2	1011		3:3 637		3:4	401		3:5	253		3:6	159
4:0	2404		4:1	1515		4:2	954			4:3 601		4:4	379		4:5	239		4:6	150
5:0	2269		5:1	1430		5:2	901			5:3 567		5:4	357		5:5	225		5:6	142
6:0	2142		6:1	1349		6:2	850			6:3 536		6:4	337		6:5	213		6:6	134
7:0	2022		7:1	1274		7:2	802			7:3 505		7:4	318		7:5	201		7:6	126  (5.051 cents error)
																			
0	3028		3029	4937		4938	6140		6141	6898		6899	7376		7377	7678		7679	7868
0	2858		2859	4660		4661	5796		5797	6512		6513	6963		6964	7248		7249	7428
0	2698		2699	4399		4400	5471		5472	6147		6148	6573		6574	6842		6843	7012
0	2546		2547	4152		4153	5164		5165	5802		5803	6204		6205	6458		6459	6618
0	2403		2404	3919		3920	4874		4875	5476		5477	5856		5857	6096		6097	6247
0	2268		2269	3699		3700	4601		4602	5169		5170	5527		5528	5753		5754	5896
0	2141		2142	3491		3492	4342		4343	4879		4880	5217		5218	5431		5432	5566
0	2021		2022	3296		3297	4099		4100	4605		4606	4924		4925	5126		5127	5253
																			
0000	0BD4		0BD5	1349		134A	17FC		17FD	1AF2		1AF3	1CD0		1CD1	1DFE		1DFF	1EBC
0000	0B2A		0B2B	1234		1235	16A4		16A5	1970		1971	1B33		1B34	1C50		1C51	1D04
0000	0A8A		0A8B	112F		1130	155F		1560	1803		1804	19AD		19AE	1ABA		1ABB	1B64
0000	09F2		09F3	1038		1039	142C		142D	16AA		16AB	183C		183D	193A		193B	19DA
0000	0963		0964	0F4F		0F50	130A		130B	1564		1565	16E0		16E1	17D0		17D1	1867
0000	08DC		08DD	0E73		0E74	11F9		11FA	1431		1432	1597		1598	1679		167A	1708
0000	085D		085E	0DA3		0DA4	10F6		10F7	130F		1310	1461		1462	1537		1538	15BE
0000	07E5		07E6	0CE0		0CE1	1003		1004	11FD		11FE	133C		133D	1406		1407	1485

*/
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// Tuning ROMs (8)

  reg [12:0] MH0StrStart;
  reg [12:0] MH0StrEnd;
  always @ ( posedge sysclk )
    begin
    case ( STR0 )
// 0000	0BD4		0BD5	1349		134A	17FC		17FD	1AF2		1AF3	1CD0		1CD1	1DFE		1DFF	1EBC
    3'h0: begin MH0StrStart <= 13'h0000; MH0StrEnd <= 13'h0BD4; end
    3'h1: begin MH0StrStart <= 13'h0BD5; MH0StrEnd <= 13'h1349; end
    3'h2: begin MH0StrStart <= 13'h134A; MH0StrEnd <= 13'h17FC; end
    3'h3: begin MH0StrStart <= 13'h17FD; MH0StrEnd <= 13'h1AF2; end
    3'h4: begin MH0StrStart <= 13'h1AF3; MH0StrEnd <= 13'h1CD0; end
    3'h5: begin MH0StrStart <= 13'h1CD1; MH0StrEnd <= 13'h1DFE; end
    3'h6: begin MH0StrStart <= 13'h1DFF; MH0StrEnd <= 13'h1EBC; end
    endcase
    end

  reg [12:0] MH1StrStart;
  reg [12:0] MH1StrEnd;
  always @ ( posedge sysclk )
    begin
    case ( STR1 )
// 0000	0B2A		0B2B	1234		1235	16A4		16A5	1970		1971	1B33		1B34	1C50		1C51	1D04
    3'h0: begin MH1StrStart <= 13'h0000; MH1StrEnd <= 13'h0B2A; end
    3'h1: begin MH1StrStart <= 13'h0B2B; MH1StrEnd <= 13'h1234; end
    3'h2: begin MH1StrStart <= 13'h1235; MH1StrEnd <= 13'h16A4; end
    3'h3: begin MH1StrStart <= 13'h16A5; MH1StrEnd <= 13'h1970; end
    3'h4: begin MH1StrStart <= 13'h1971; MH1StrEnd <= 13'h1B33; end
    3'h5: begin MH1StrStart <= 13'h1B34; MH1StrEnd <= 13'h1C50; end
    3'h6: begin MH1StrStart <= 13'h1C51; MH1StrEnd <= 13'h1D04; end
    endcase
    end

  reg [12:0] MH2StrStart;
  reg [12:0] MH2StrEnd;
  always @ ( posedge sysclk )
    begin
    case ( STR2 )
// 0000	0A8A		0A8B	112F		1130	155F		1560	1803		1804	19AD		19AE	1ABA		1ABB	1B64
    3'h0: begin MH2StrStart <= 13'h0000; MH2StrEnd <= 13'h0A8A; end
    3'h1: begin MH2StrStart <= 13'h0A8B; MH2StrEnd <= 13'h112F; end
    3'h2: begin MH2StrStart <= 13'h1130; MH2StrEnd <= 13'h155F; end
    3'h3: begin MH2StrStart <= 13'h1560; MH2StrEnd <= 13'h1803; end
    3'h4: begin MH2StrStart <= 13'h1804; MH2StrEnd <= 13'h19AD; end
    3'h5: begin MH2StrStart <= 13'h19AE; MH2StrEnd <= 13'h1ABA; end
    3'h6: begin MH2StrStart <= 13'h1ABB; MH2StrEnd <= 13'h1B64; end
    endcase
    end

  reg [12:0] MH3StrStart;
  reg [12:0] MH3StrEnd;
  always @ ( posedge sysclk )
    begin
    case ( STR3 )
// 0000	09F2		09F3	1038		1039	142C		142D	16AA		16AB	183C		183D	193A		193B	19DA
    3'h0: begin MH3StrStart <= 13'h0000; MH3StrEnd <= 13'h09F2; end
    3'h1: begin MH3StrStart <= 13'h09F3; MH3StrEnd <= 13'h1038; end
    3'h2: begin MH3StrStart <= 13'h1039; MH3StrEnd <= 13'h142C; end
    3'h3: begin MH3StrStart <= 13'h142D; MH3StrEnd <= 13'h16AA; end
    3'h4: begin MH3StrStart <= 13'h16AB; MH3StrEnd <= 13'h183C; end
    3'h5: begin MH3StrStart <= 13'h183D; MH3StrEnd <= 13'h193A; end
    3'h6: begin MH3StrStart <= 13'h193B; MH3StrEnd <= 13'h19DA; end
    endcase
    end

  reg [12:0] MH4StrStart;
  reg [12:0] MH4StrEnd;
  always @ ( posedge sysclk )
    begin
    case ( STR4 )
// 0000	0963		0964	0F4F		0F50	130A		130B	1564		1565	16E0		16E1	17D0		17D1	1867
    3'h0: begin MH4StrStart <= 13'h0000; MH4StrEnd <= 13'h0963; end
    3'h1: begin MH4StrStart <= 13'h0964; MH4StrEnd <= 13'h0F4F; end
    3'h2: begin MH4StrStart <= 13'h0F50; MH4StrEnd <= 13'h130A; end
    3'h3: begin MH4StrStart <= 13'h130B; MH4StrEnd <= 13'h1564; end
    3'h4: begin MH4StrStart <= 13'h1565; MH4StrEnd <= 13'h16E0; end
    3'h5: begin MH4StrStart <= 13'h16E1; MH4StrEnd <= 13'h17D0; end
    3'h6: begin MH4StrStart <= 13'h17D1; MH4StrEnd <= 13'h1867; end
    endcase
    end

  reg [12:0] MH5StrStart;
  reg [12:0] MH5StrEnd;
  always @ ( posedge sysclk )
    begin
    case ( STR5 )
// 0000	08DC		08DD	0E73		0E74	11F9		11FA	1431		1432	1597		1598	1679		167A	1708
    3'h0: begin MH5StrStart <= 13'h0000; MH5StrEnd <= 13'h08DC; end
    3'h1: begin MH5StrStart <= 13'h08DD; MH5StrEnd <= 13'h0E73; end
    3'h2: begin MH5StrStart <= 13'h0E74; MH5StrEnd <= 13'h11F9; end
    3'h3: begin MH5StrStart <= 13'h11FA; MH5StrEnd <= 13'h1431; end
    3'h4: begin MH5StrStart <= 13'h1432; MH5StrEnd <= 13'h1597; end
    3'h5: begin MH5StrStart <= 13'h1598; MH5StrEnd <= 13'h1679; end
    3'h6: begin MH5StrStart <= 13'h167A; MH5StrEnd <= 13'h1708; end
    endcase
    end

  reg [12:0] MH6StrStart;
  reg [12:0] MH6StrEnd;
  always @ ( posedge sysclk )
    begin
    case ( STR6 )
// 0000	085D		085E	0DA3		0DA4	10F6		10F7	130F		1310	1461		1462	1537		1538	15BE
    3'h0: begin MH6StrStart <= 13'h0000; MH6StrEnd <= 13'h085D; end
    3'h1: begin MH6StrStart <= 13'h085E; MH6StrEnd <= 13'h0DA3; end
    3'h2: begin MH6StrStart <= 13'h0DA4; MH6StrEnd <= 13'h10F6; end
    3'h3: begin MH6StrStart <= 13'h10F7; MH6StrEnd <= 13'h130F; end
    3'h4: begin MH6StrStart <= 13'h1310; MH6StrEnd <= 13'h1461; end
    3'h5: begin MH6StrStart <= 13'h1462; MH6StrEnd <= 13'h1537; end
    3'h6: begin MH6StrStart <= 13'h1538; MH6StrEnd <= 13'h15BE; end
    endcase
    end

  reg [12:0] MH7StrStart;
  reg [12:0] MH7StrEnd;
  always @ ( posedge sysclk )
    begin
    case ( STR7 )
// 0000	07E5		07E6	0CE0		0CE1	1003		1004	11FD		11FE	133C		133D	1406		1407	1407
    3'h0: begin MH7StrStart <= 13'h0000; MH7StrEnd <= 13'h07E5; end
    3'h1: begin MH7StrStart <= 13'h07E6; MH7StrEnd <= 13'h0CE0; end
    3'h2: begin MH7StrStart <= 13'h0CE1; MH7StrEnd <= 13'h1003; end
    3'h3: begin MH7StrStart <= 13'h1004; MH7StrEnd <= 13'h11FD; end
    3'h4: begin MH7StrStart <= 13'h11FE; MH7StrEnd <= 13'h133C; end
    3'h5: begin MH7StrStart <= 13'h133D; MH7StrEnd <= 13'h1406; end
    3'h6: begin MH7StrStart <= 13'h1407; MH7StrEnd <= 13'h1485; end  // This string has 5.051 cents error
    endcase
    end

////////////////////////////////////////////////////////////////////////////////////////////////////////////
// String RAMs (delay lines)

  reg [17:0]  SDL0in = 18'h0000;             // register holds data for next RAM write
  reg [17:0]  SDL1in = 18'h0000;             // register holds data for next RAM write
  reg [17:0]  SDL2in = 18'h0000;             // register holds data for next RAM write
  reg [17:0]  SDL3in = 18'h0000;             // register holds data for next RAM write
  reg [17:0]  SDL4in = 18'h0000;             // register holds data for next RAM write
  reg [17:0]  SDL5in = 18'h0000;             // register holds data for next RAM write
  reg [17:0]  SDL6in = 18'h0000;             // register holds data for next RAM write
  reg [17:0]  SDL7in = 18'h0000;             // register holds data for next RAM write
  
  reg [17:0]  SDL0out = 18'h0000;            // RAM output
  reg [17:0]  SDL1out = 18'h0000;            // RAM output
  reg [17:0]  SDL2out = 18'h0000;            // RAM output
  reg [17:0]  SDL3out = 18'h0000;            // RAM output
  reg [17:0]  SDL4out = 18'h0000;            // RAM output
  reg [17:0]  SDL5out = 18'h0000;            // RAM output
  reg [17:0]  SDL6out = 18'h0000;            // RAM output
  reg [17:0]  SDL7out = 18'h0000;            // RAM output
  
  reg [12:0]  WG0adr = 0;                        // state machine waveguide address register
  reg [12:0]  WG1adr = 0;                        // state machine waveguide address register
  reg [12:0]  WG2adr = 0;                        // state machine waveguide address register
  reg [12:0]  WG3adr = 0;                        // state machine waveguide address register
  reg [12:0]  WG4adr = 0;                        // state machine waveguide address register
  reg [12:0]  WG5adr = 0;                        // state machine waveguide address register
  reg [12:0]  WG6adr = 0;                        // state machine waveguide address register
  reg [12:0]  WG7adr = 0;                        // state machine waveguide address register

// RAMs to store each delay line address
  reg [12:0]  SDL0adr[7:0];                    // DIST RAM
  reg [12:0]  SDL1adr[7:0];
  reg [12:0]  SDL2adr[7:0];
  reg [12:0]  SDL3adr[7:0];
  reg [12:0]  SDL4adr[7:0];
  reg [12:0]  SDL5adr[7:0];
  reg [12:0]  SDL6adr[7:0];
  reg [12:0]  SDL7adr[7:0];
  
  // instantiate 8 mini-harp RAMs:  
  reg signed [17:0] MH0RAM [8191:0];         // BLOCK RAM  (9 BRAMs per 8K x 18 bits)
  reg signed [17:0] MH1RAM [8191:0];
  reg signed [17:0] MH2RAM [8191:0];
  reg signed [17:0] MH3RAM [8191:0];
  reg signed [17:0] MH4RAM [8191:0];
  reg signed [17:0] MH5RAM [8191:0];
  reg signed [17:0] MH6RAM [8191:0];
  reg signed [17:0] MH7RAM [8191:0];
 
  reg RAMWRT = 1'b0;

  always @ ( posedge sysclk )
    begin
    SDL0out <= MH0RAM[WG0adr];
    SDL1out <= MH1RAM[WG1adr];
    SDL2out <= MH2RAM[WG2adr];
    SDL3out <= MH3RAM[WG3adr];
    SDL4out <= MH4RAM[WG4adr];
    SDL5out <= MH5RAM[WG5adr];
    SDL6out <= MH6RAM[WG6adr];
    SDL7out <= MH7RAM[WG7adr];
    if ( RAMWRT )
      begin
      MH0RAM[WG0adr] <= SDL0in;
      MH1RAM[WG1adr] <= SDL1in;
      MH2RAM[WG2adr] <= SDL2in;
      MH3RAM[WG3adr] <= SDL3in;
      MH4RAM[WG4adr] <= SDL4in;
      MH5RAM[WG5adr] <= SDL5in;
      MH6RAM[WG6adr] <= SDL6in;
      MH7RAM[WG7adr] <= SDL7in;
      end
    end
  
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// Reflection point filter -  a single filter is shared between all 8 strings.
  reg                IIRena = 1'b0;

  reg  signed [17:0] IIR0in = 0;          // IIR0in is input register for IIR0, also serves as string loop output
  reg  signed [17:0] IIR1in = 0;          // IIR1in is input register for IIR1, also serves as string loop output
  reg  signed [17:0] IIR2in = 0;          // IIR2in is input register for IIR2, also serves as string loop output
  reg  signed [17:0] IIR3in = 0;          // IIR3in is input register for IIR3, also serves as string loop output
  reg  signed [17:0] IIR4in = 0;          // IIR4in is input register for IIR4, also serves as string loop output
  reg  signed [17:0] IIR5in = 0;          // IIR5in is input register for IIR5, also serves as string loop output
  reg  signed [17:0] IIR6in = 0;          // IIR6in is input register for IIR6, also serves as string loop output
  reg  signed [17:0] IIR7in = 0;          // IIR7in is input register for IIR7, also serves as string loop output

  wire signed [17:0] IIR0out;         // output of filter
  wire signed [17:0] IIR1out;         // output of filter
  wire signed [17:0] IIR2out;         // output of filter
  wire signed [17:0] IIR3out;         // output of filter
  wire signed [17:0] IIR4out;         // output of filter
  wire signed [17:0] IIR5out;         // output of filter
  wire signed [17:0] IIR6out;         // output of filter
  wire signed [17:0] IIR7out;         // output of filter
 
  reg  signed [17:0] IIR0outREG = 0;
  reg  signed [17:0] IIR1outREG = 0;
  reg  signed [17:0] IIR2outREG = 0;
  reg  signed [17:0] IIR3outREG = 0;
  reg  signed [17:0] IIR4outREG = 0;
  reg  signed [17:0] IIR5outREG = 0;
  reg  signed [17:0] IIR6outREG = 0;
  reg  signed [17:0] IIR7outREG = 0;
  
  reg signed [17:0] BW0 = 0;             // bandwidth
  reg signed [17:0] BW1 = 0;             // bandwidth
  reg signed [17:0] BW2 = 0;             // bandwidth
  reg signed [17:0] BW3 = 0;             // bandwidth
  reg signed [17:0] BW4 = 0;             // bandwidth
  reg signed [17:0] BW5 = 0;             // bandwidth
  reg signed [17:0] BW6 = 0;             // bandwidth
  reg signed [17:0] BW7 = 0;             // bandwidth

  IIRnew IIR0 ( .clk( sysclk ), .ena( IIRena ), .I( IIR0in ), .BW( BW0 ), .SEL( STR0 ), .O( IIR0out ) );
  IIRnew IIR1 ( .clk( sysclk ), .ena( IIRena ), .I( IIR1in ), .BW( BW1 ), .SEL( STR1 ), .O( IIR1out ) );
  IIRnew IIR2 ( .clk( sysclk ), .ena( IIRena ), .I( IIR2in ), .BW( BW2 ), .SEL( STR2 ), .O( IIR2out ) );
  IIRnew IIR3 ( .clk( sysclk ), .ena( IIRena ), .I( IIR3in ), .BW( BW3 ), .SEL( STR3 ), .O( IIR3out ) );
  IIRnew IIR4 ( .clk( sysclk ), .ena( IIRena ), .I( IIR4in ), .BW( BW4 ), .SEL( STR4 ), .O( IIR4out ) );
  IIRnew IIR5 ( .clk( sysclk ), .ena( IIRena ), .I( IIR5in ), .BW( BW5 ), .SEL( STR5 ), .O( IIR5out ) );
  IIRnew IIR6 ( .clk( sysclk ), .ena( IIRena ), .I( IIR6in ), .BW( BW6 ), .SEL( STR6 ), .O( IIR6out ) );
  IIRnew IIR7 ( .clk( sysclk ), .ena( IIRena ), .I( IIR7in ), .BW( BW7 ), .SEL( STR7 ), .O( IIR7out ) );
	 
//////////////////////////////////////////////////////////////////////////////////////////////////
// for EXval:

  reg signed [17:0] EX0val[7:0];         // waveguide EXval (excitation value) signal;
  reg signed [17:0] EX1val[7:0];         // waveguide EXval (excitation value) signal;   DIST RAM
  reg signed [17:0] EX2val[7:0];         // waveguide EXval (excitation value) signal;
  reg signed [17:0] EX3val[7:0];         // waveguide EXval (excitation value) signal;
  reg signed [17:0] EX4val[7:0];         // waveguide EXval (excitation value) signal;
  reg signed [17:0] EX5val[7:0];         // waveguide EXval (excitation value) signal;
  reg signed [17:0] EX6val[7:0];         // waveguide EXval (excitation value) signal;
  reg signed [17:0] EX7val[7:0];         // waveguide EXval (excitation value) signal;
  
  reg [7:0] EX0 = 0;               // 8 flag bits - hi = excite active
  reg [7:0] EX1 = 0;               // 8 flag bits - hi = excite active
  reg [7:0] EX2 = 0;               // 8 flag bits - hi = excite active
  reg [7:0] EX3 = 0;               // 8 flag bits - hi = excite active
  reg [7:0] EX4 = 0;               // 8 flag bits - hi = excite active
  reg [7:0] EX5 = 0;               // 8 flag bits - hi = excite active
  reg [7:0] EX6 = 0;               // 8 flag bits - hi = excite active
  reg [7:0] EX7 = 0;               // 8 flag bits - hi = excite active

  reg [12:0] EX0cnt[7:0];          // EXval pulse width counter  DIST RAM
  reg [12:0] EX1cnt[7:0];          // EXval pulse width counter
  reg [12:0] EX2cnt[7:0];          // EXval pulse width counter
  reg [12:0] EX3cnt[7:0];          // EXval pulse width counter
  reg [12:0] EX4cnt[7:0];          // EXval pulse width counter
  reg [12:0] EX5cnt[7:0];          // EXval pulse width counter
  reg [12:0] EX6cnt[7:0];          // EXval pulse width counter
  reg [12:0] EX7cnt[7:0];          // EXval pulse width counter

  reg signed [17:0] EX0valtmp = 0;       // temp reg for computing EXval[] per string
  reg signed [17:0] EX1valtmp = 0;       // temp reg for computing EXval[] per string
  reg signed [17:0] EX2valtmp = 0;       // temp reg for computing EXval[] per string
  reg signed [17:0] EX3valtmp = 0;       // temp reg for computing EXval[] per string
  reg signed [17:0] EX4valtmp = 0;       // temp reg for computing EXval[] per string
  reg signed [17:0] EX5valtmp = 0;       // temp reg for computing EXval[] per string
  reg signed [17:0] EX6valtmp = 0;       // temp reg for computing EXval[] per string
  reg signed [17:0] EX7valtmp = 0;       // temp reg for computing EXval[] per string

  reg [12:0] EX0cnttmp = 0;              // temp reg for computing EXcnt[] per string
  reg [12:0] EX1cnttmp = 0;              // temp reg for computing EXcnt[] per string
  reg [12:0] EX2cnttmp = 0;              // temp reg for computing EXcnt[] per string
  reg [12:0] EX3cnttmp = 0;              // temp reg for computing EXcnt[] per string
  reg [12:0] EX4cnttmp = 0;              // temp reg for computing EXcnt[] per string
  reg [12:0] EX5cnttmp = 0;              // temp reg for computing EXcnt[] per string
  reg [12:0] EX6cnttmp = 0;              // temp reg for computing EXcnt[] per string
  reg [12:0] EX7cnttmp = 0;              // temp reg for computing EXcnt[] per string

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// Waveguide state machine

// 8 sign bits for each mini-harp
  reg [7:0] sign0 = 0;              // used to invert sign for each excite to prevent "charging" the system.
  reg [7:0] sign1 = 0;              // used to invert sign for each excite to prevent "charging" the system.
  reg [7:0] sign2 = 0;              // used to invert sign for each excite to prevent "charging" the system.
  reg [7:0] sign3 = 0;              // used to invert sign for each excite to prevent "charging" the system.
  reg [7:0] sign4 = 0;              // used to invert sign for each excite to prevent "charging" the system.
  reg [7:0] sign5 = 0;              // used to invert sign for each excite to prevent "charging" the system.
  reg [7:0] sign6 = 0;              // used to invert sign for each excite to prevent "charging" the system.
  reg [7:0] sign7 = 0;              // used to invert sign for each excite to prevent "charging" the system.

  wire signed [23:0] DACbus0;
  wire signed [23:0] DACbus1;
  wire signed [23:0] DACbus2;
  wire signed [23:0] DACbus3;
  wire signed [23:0] DACbus4;
  wire signed [23:0] DACbus5;

// mini harp output registers:  (these are accumulator registers for each mini harp)
  reg signed [20:0] MH0sum = 0;
  reg signed [20:0] MH1sum = 0;
  reg signed [20:0] MH2sum = 0;
  reg signed [20:0] MH3sum = 0;
  reg signed [20:0] MH4sum = 0;
  reg signed [20:0] MH5sum = 0;
  reg signed [20:0] MH6sum = 0;
  reg signed [20:0] MH7sum = 0;

  assign DACbus0 = {{3{MH0sum[20]}},MH0sum} + {{3{MH1sum[20]}},MH1sum};
  assign DACbus1 = {{3{MH2sum[20]}},MH2sum} + {{3{MH3sum[20]}},MH3sum};
  assign DACbus2 = {{3{MH4sum[20]}},MH4sum} + {{3{MH5sum[20]}},MH5sum};
  assign DACbus3 = {{3{MH6sum[20]}},MH6sum} + {{3{MH7sum[20]}},MH7sum};

  assign DACbus4 = DACbus0 + DACbus1;
  assign DACbus5 = DACbus2 + DACbus3;

  assign DACbus = DACbus4 + DACbus5;

//////////////////////////////////////////////////////////////////////////////////////////////////

//  wire signed [17:0] WGin;                       // Waveguide inputs
//  assign WGin = ( sign[STR] == 1'b1 ) ? (-IIRoutREG + EXvaltmp) : (-IIRoutREG - EXvaltmp) ;
  reg signed [17:0] WG0in = 0;
  reg signed [17:0] WG1in = 0;
  reg signed [17:0] WG2in = 0;
  reg signed [17:0] WG3in = 0;
  reg signed [17:0] WG4in = 0;
  reg signed [17:0] WG5in = 0;
  reg signed [17:0] WG6in = 0;
  reg signed [17:0] WG7in = 0;

  reg signed [17:0] WG0inP = 0;
  reg signed [17:0] WG1inP = 0;
  reg signed [17:0] WG2inP = 0;
  reg signed [17:0] WG3inP = 0;
  reg signed [17:0] WG4inP = 0;
  reg signed [17:0] WG5inP = 0;
  reg signed [17:0] WG6inP = 0;
  reg signed [17:0] WG7inP = 0;

  reg signed [17:0] WG0inN = 0;
  reg signed [17:0] WG1inN = 0;
  reg signed [17:0] WG2inN = 0;
  reg signed [17:0] WG3inN = 0;
  reg signed [17:0] WG4inN = 0;
  reg signed [17:0] WG5inN = 0;
  reg signed [17:0] WG6inN = 0;
  reg signed [17:0] WG7inN = 0;

// overflow detection:
  wire oflo0;
  wire oflo1;
  wire oflo2;
  wire oflo3;
  wire oflo4;
  wire oflo5;
  wire oflo6;
  wire oflo7;

  assign oflo0 = (( EX0valtmp[17] ^ IIR0outREG[17] ) & ( WG0in[17] ^ EX0valtmp[17] ));
  assign oflo1 = (( EX1valtmp[17] ^ IIR1outREG[17] ) & ( WG1in[17] ^ EX1valtmp[17] ));
  assign oflo2 = (( EX2valtmp[17] ^ IIR2outREG[17] ) & ( WG2in[17] ^ EX2valtmp[17] ));
  assign oflo3 = (( EX3valtmp[17] ^ IIR3outREG[17] ) & ( WG3in[17] ^ EX3valtmp[17] ));
  assign oflo4 = (( EX4valtmp[17] ^ IIR4outREG[17] ) & ( WG4in[17] ^ EX4valtmp[17] ));
  assign oflo5 = (( EX5valtmp[17] ^ IIR5outREG[17] ) & ( WG5in[17] ^ EX5valtmp[17] ));
  assign oflo6 = (( EX6valtmp[17] ^ IIR6outREG[17] ) & ( WG6in[17] ^ EX6valtmp[17] ));
  assign oflo7 = (( EX7valtmp[17] ^ IIR7outREG[17] ) & ( WG7in[17] ^ EX7valtmp[17] ));

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
  reg [12:0] WG0len = 0;
  reg [12:0] WG1len = 0;
  reg [12:0] WG2len = 0;
  reg [12:0] WG3len = 0;
  reg [12:0] WG4len = 0;
  reg [12:0] WG5len = 0;
  reg [12:0] WG6len = 0;
  reg [12:0] WG7len = 0;

// state machine structure:
  reg [3:0] state = 4'h0;
  reg run = 1'b0;
  reg pass = 0;                               // state machine pass register for decimation 2:1
  
  always @ ( posedge sysclk )
    begin
    if ( DACena )
      begin
      state <= 4'h0;                            // starting state
      run <= 1'b1;                              // tell state machine to run
      pass <= SW[7];                            // SW[7] = 1 for low range, 0 for high.
      DACreg <= DACbus;                         // write current DACbus value to DACreg for output
      end
    
    else        // else if NOT DACena:
      
      begin
      if ( run )
        begin
        case ( state ) 
        4'h0:
           begin  state <= state + 1;
           MH0sum <= 0;
           MH1sum <= 0;
           MH2sum <= 0;
           MH3sum <= 0;
           MH4sum <= 0;
           MH5sum <= 0;
           MH6sum <= 0;
           MH7sum <= 0;
           STR <= 3'h0;                              // Select string zero to start
           end
        
        4'h1:
          begin  state <= state + 1;
          STR0 <= STR;
          STR1 <= STR;
          STR2 <= STR;
          STR3 <= STR;
          STR4 <= STR;
          STR5 <= STR;
          STR6 <= STR;
          STR7 <= STR;
          end
        
        4'h2:                        // initialize for this string
          begin  state <= state + 1;
          WG0adr <= SDL0adr[STR0];
          WG1adr <= SDL1adr[STR1];
          WG2adr <= SDL2adr[STR2];
          WG3adr <= SDL3adr[STR3];
          WG4adr <= SDL4adr[STR4];
          WG5adr <= SDL5adr[STR5];
          WG6adr <= SDL6adr[STR6];
          WG7adr <= SDL7adr[STR7];
          end

        4'h3:                        
          begin  state <= state + 1;

          WG0len <= ((MH0StrEnd - MH0StrStart + 1)>>>SW[6:4]);
          WG1len <= ((MH1StrEnd - MH1StrStart + 1)>>>SW[6:4]);
          WG2len <= ((MH2StrEnd - MH2StrStart + 1)>>>SW[6:4]);
          WG3len <= ((MH3StrEnd - MH3StrStart + 1)>>>SW[6:4]);
          WG4len <= ((MH4StrEnd - MH4StrStart + 1)>>>SW[6:4]);
          WG5len <= ((MH5StrEnd - MH5StrStart + 1)>>>SW[6:4]);
          WG6len <= ((MH6StrEnd - MH6StrStart + 1)>>>SW[6:4]);
          WG7len <= ((MH7StrEnd - MH7StrStart + 1)>>>SW[6:4]);

          // basic keyboard tracking - may be eliminated by configured BW for each string (later version)

          BW0 <= 18'sh01000 + ( ( {STR,3'h0} + 1 ) << MOD_WHL[6:3] );
          BW1 <= 18'sh01000 + ( ( {STR,3'h1} + 1 ) << MOD_WHL[6:3] );
          BW2 <= 18'sh01000 + ( ( {STR,3'h2} + 1 ) << MOD_WHL[6:3] );
          BW3 <= 18'sh01000 + ( ( {STR,3'h3} + 1 ) << MOD_WHL[6:3] );
          BW4 <= 18'sh01000 + ( ( {STR,3'h4} + 1 ) << MOD_WHL[6:3] );
          BW5 <= 18'sh01000 + ( ( {STR,3'h5} + 1 ) << MOD_WHL[6:3] );
          BW6 <= 18'sh01000 + ( ( {STR,3'h6} + 1 ) << MOD_WHL[6:3] );
          BW7 <= 18'sh01000 + ( ( {STR,3'h7} + 1 ) << MOD_WHL[6:3] );

          end

        4'h4:
          begin  state <= state + 1;
          if ( PB[3] | PB[2] | PB[1] | ~SUSTAIN )
            begin

            IIR0in <= 0;  // crude, but effective.
            IIR1in <= 0;
            IIR2in <= 0;
            IIR3in <= 0;
            IIR4in <= 0;
            IIR5in <= 0;
            IIR6in <= 0;
            IIR7in <= 0;

/* 
//This didn't work - It squelches by lifting the sustain pedal, but when it is then pressed, the 
//same waveguides are resonating again.
            IIR0in <= SDL0out >>> 1;            // transfer current string waveguide output to IIR input register
            IIR1in <= SDL1out >>> 1;            // transfer current string waveguide output to IIR input register
            IIR2in <= SDL2out >>> 1;            // transfer current string waveguide output to IIR input register
            IIR3in <= SDL3out >>> 1;            // transfer current string waveguide output to IIR input register
            IIR4in <= SDL4out >>> 1;            // transfer current string waveguide output to IIR input register
            IIR5in <= SDL5out >>> 1;            // transfer current string waveguide output to IIR input register
            IIR6in <= SDL6out >>> 1;            // transfer current string waveguide output to IIR input register
            IIR7in <= SDL7out >>> 1;            // transfer current string waveguide output to IIR input register
*/            
            end
          else
            begin
            IIR0in <= SDL0out;            // transfer current string waveguide output to IIR input register
            IIR1in <= SDL1out;            // transfer current string waveguide output to IIR input register
            IIR2in <= SDL2out;            // transfer current string waveguide output to IIR input register
            IIR3in <= SDL3out;            // transfer current string waveguide output to IIR input register
            IIR4in <= SDL4out;            // transfer current string waveguide output to IIR input register
            IIR5in <= SDL5out;            // transfer current string waveguide output to IIR input register
            IIR6in <= SDL6out;            // transfer current string waveguide output to IIR input register
            IIR7in <= SDL7out;            // transfer current string waveguide output to IIR input register
            end
            
          IIRena <= 1'b1;             // tell filter to go, takes 3 clocks          
          end

        4'h5:                         // filter cycle 1 complete
          begin  state <= state + 1;

          IIRena <= 1'b0;             // deassert the filter go flag
          
          EX0valtmp <= EX0val[STR0];    // sets default value for EXvaltmp in case next state's logic doesn't change it.
          EX1valtmp <= EX1val[STR1];
          EX2valtmp <= EX2val[STR2];
          EX3valtmp <= EX3val[STR3];
          EX4valtmp <= EX4val[STR4];
          EX5valtmp <= EX5val[STR5];
          EX6valtmp <= EX6val[STR6];
          EX7valtmp <= EX7val[STR7];

          EX0cnttmp <= EX0cnt[STR0];
          EX1cnttmp <= EX1cnt[STR1];
          EX2cnttmp <= EX2cnt[STR2];
          EX3cnttmp <= EX3cnt[STR3];
          EX4cnttmp <= EX4cnt[STR4];
          EX5cnttmp <= EX5cnt[STR5];
          EX6cnttmp <= EX6cnt[STR6];
          EX7cnttmp <= EX7cnt[STR7];
          end

        4'h6:                         // filter cycle 2 complete
          begin  state <= state + 1;
          if ( Pluck == 1'b1 )
            begin


            if ( Plucked_Str == STR )
              begin

              Clr_Pluck <= 1'b1;             // A pluck event cannot occur more than once per sample time
                                             // so we can simply clear it once we see it.

              case ( Plucked_MH )

              4'h0:
                begin              
                EX0[STR0]   <= 1'b1;                  // Set excite flag to start pulse
                sign0[STR0] <= ~sign0[STR0];          // alternates which way the excite signal is signed.
                EX0cnttmp  <= WG0len ;                // length is calculated in state 0
                EX0valtmp  <= ( 18'h1FFFF ) ;         // velocity disabled this version
                end

              4'h1:
                begin              
                EX1[STR1]   <= 1'b1;                  // Set excite flag to start pulse
                sign1[STR1] <= ~sign1[STR1];          // alternates which way the excite signal is signed.
                EX1cnttmp  <= WG1len ;                // length is calculated in state 0
                EX1valtmp  <= ( 18'h1FFFF ) ;         // velocity disabled this version
                end

              4'h2:
                begin              
                EX2[STR2]   <= 1'b1;                  // Set excite flag to start pulse
                sign2[STR2] <= ~sign2[STR2];          // alternates which way the excite signal is signed.
                EX2cnttmp  <= WG2len ;                // length is calculated in state 0
                EX2valtmp  <= ( 18'h1FFFF ) ;         // velocity disabled this version
                end

              4'h3:
                begin              
                EX3[STR3]   <= 1'b1;                  // Set excite flag to start pulse
                sign3[STR3] <= ~sign3[STR3];          // alternates which way the excite signal is signed.
                EX3cnttmp  <= WG3len ;                // length is calculated in state 0
                EX3valtmp  <= ( 18'h1FFFF ) ;         // velocity disabled this version
                end

              4'h4:
                begin              
                EX4[STR4]   <= 1'b1;                  // Set excite flag to start pulse
                sign4[STR4] <= ~sign4[STR4];          // alternates which way the excite signal is signed.
                EX4cnttmp  <= WG4len ;                // length is calculated in state 0
                EX4valtmp  <= ( 18'h1FFFF ) ;         // velocity disabled this version
                end

              4'h5:
                begin              
                EX5[STR5]   <= 1'b1;                  // Set excite flag to start pulse
                sign5[STR5] <= ~sign5[STR5];          // alternates which way the excite signal is signed.
                EX5cnttmp  <= WG5len ;                // length is calculated in state 0
                EX5valtmp  <= ( 18'h1FFFF ) ;         // velocity disabled this version
                end

              4'h6:
                begin              
                EX6[STR6]   <= 1'b1;                  // Set excite flag to start pulse
                sign6[STR6] <= ~sign6[STR6];          // alternates which way the excite signal is signed.
                EX6cnttmp  <= WG6len ;                // length is calculated in state 0
                EX6valtmp  <= ( 18'h1FFFF ) ;         // velocity disabled this version
                end


              4'h7:
                begin
                EX7[STR7]   <= 1'b1;                  // Set excite flag to start pulse
                sign7[STR7] <= ~sign7[STR7];          // alternates which way the excite signal is signed.
                EX7cnttmp  <= WG7len ;                // length is calculated in state 0
                EX7valtmp  <= ( 18'h1FFFF ) ;         // velocity disabled this version
                end

              endcase
              end
            end
          end            

        4'h7:
          begin  state <= state + 1;
          
          if ( EX0[STR0] == 1'b1 )              // if this string is currently being excited:
            begin
            EX0cnttmp <= EX0cnttmp - 13'h0001;  // for rectangular pulse
            if ( EX0cnttmp == 13'h0000 )
              begin
              EX0valtmp <= 18'h00000;           // return value to zero
              EX0[STR0]  <= 1'b0;               // turn off flag excite to end pulse
              end
            end

          if ( EX1[STR1] == 1'b1 )              // if this string is currently being excited:
            begin
            EX1cnttmp <= EX1cnttmp - 13'h0001;  // for rectangular pulse
            if ( EX1cnttmp == 13'h0000 )
              begin
              EX1valtmp <= 18'h00000;           // return value to zero
              EX1[STR1]  <= 1'b0;               // turn off flag excite to end pulse
              end
            end

          if ( EX2[STR2] == 1'b1 )              // if this string is currently being excited:
            begin
            EX2cnttmp <= EX2cnttmp - 13'h0001;  // for rectangular pulse
            if ( EX2cnttmp == 13'h0000 )
              begin
              EX2valtmp <= 18'h00000;           // return value to zero
              EX2[STR2]  <= 1'b0;               // turn off flag excite to end pulse
              end
            end

          if ( EX3[STR3] == 1'b1 )              // if this string is currently being excited:
            begin
            EX3cnttmp <= EX3cnttmp - 13'h0001;  // for rectangular pulse
            if ( EX3cnttmp == 13'h0000 )
              begin
              EX3valtmp <= 18'h00000;           // return value to zero
              EX3[STR3]  <= 1'b0;               // turn off flag excite to end pulse
              end
            end

          if ( EX4[STR4] == 1'b1 )              // if this string is currently being excited:
            begin
            EX4cnttmp <= EX4cnttmp - 13'h0001;  // for rectangular pulse
            if ( EX4cnttmp == 13'h0000 )
              begin
              EX4valtmp <= 18'h00000;           // return value to zero
              EX4[STR4]  <= 1'b0;               // turn off flag excite to end pulse
              end
            end

          if ( EX5[STR5] == 1'b1 )              // if this string is currently being excited:
            begin
            EX5cnttmp <= EX5cnttmp - 13'h0001;  // for rectangular pulse
            if ( EX5cnttmp == 13'h0000 )
              begin
              EX5valtmp <= 18'h00000;           // return value to zero
              EX5[STR5]  <= 1'b0;               // turn off flag excite to end pulse
              end
            end

          if ( EX6[STR6] == 1'b1 )              // if this string is currently being excited:
            begin
            EX6cnttmp <= EX6cnttmp - 13'h0001;  // for rectangular pulse
            if ( EX6cnttmp == 13'h0000 )
              begin
              EX6valtmp <= 18'h00000;           // return value to zero
              EX6[STR6]  <= 1'b0;               // turn off flag excite to end pulse
              end
            end

          if ( EX7[STR7] == 1'b1 )              // if this string is currently being excited:
            begin
            EX7cnttmp <= EX7cnttmp - 13'h0001;  // for rectangular pulse
            if ( EX7cnttmp == 13'h0000 )
              begin
              EX7valtmp <= 18'h00000;           // return value to zero
              EX7[STR7]  <= 1'b0;               // turn off flag excite to end pulse
              end
            end
                
          end

        4'h8:                         // filter cycle 3 complete.  IIRout is valid
          begin  state <= state + 1;
          IIR0outREG <= IIR0out;
          IIR1outREG <= IIR1out;
          IIR2outREG <= IIR2out;
          IIR3outREG <= IIR3out;
          IIR4outREG <= IIR4out;
          IIR5outREG <= IIR5out;
          IIR6outREG <= IIR6out;
          IIR7outREG <= IIR7out;
            
          Clr_Pluck <= 1'b0;          // always deassert this
          end

        4'h9:
          begin  state <= state + 1;
          
          WG0inP <= -IIR0outREG + EX0valtmp;
          WG0inN <= -IIR0outREG - EX0valtmp; 

          WG1inP <= -IIR1outREG + EX1valtmp;
          WG1inN <= -IIR1outREG - EX1valtmp; 

          WG2inP <= -IIR2outREG + EX2valtmp;
          WG2inN <= -IIR2outREG - EX2valtmp; 

          WG3inP <= -IIR3outREG + EX3valtmp;
          WG3inN <= -IIR3outREG - EX3valtmp; 

          WG4inP <= -IIR4outREG + EX4valtmp;
          WG4inN <= -IIR4outREG - EX4valtmp; 

          WG5inP <= -IIR5outREG + EX5valtmp;
          WG5inN <= -IIR5outREG - EX5valtmp; 

          WG6inP <= -IIR6outREG + EX6valtmp;
          WG6inN <= -IIR6outREG - EX6valtmp; 

          WG7inP <= -IIR7outREG + EX7valtmp;
          WG7inN <= -IIR7outREG - EX7valtmp; 

          end

        4'hA:
          begin  state <= state + 1;
          WG0in <= ( sign0[STR0] ) ? WG0inP : WG0inN ;
          WG1in <= ( sign1[STR1] ) ? WG1inP : WG1inN ;
          WG2in <= ( sign2[STR2] ) ? WG2inP : WG2inN ;
          WG3in <= ( sign3[STR3] ) ? WG3inP : WG3inN ;
          WG4in <= ( sign4[STR4] ) ? WG4inP : WG4inN ;
          WG5in <= ( sign5[STR5] ) ? WG5inP : WG5inN ;
          WG6in <= ( sign6[STR6] ) ? WG6inP : WG6inN ;
          WG7in <= ( sign7[STR7] ) ? WG7inP : WG7inN ;
          end

        4'hB:
          begin  state <= state + 1;
          EX0val[STR0] <= EX0valtmp;     // save exciter values
          EX1val[STR1] <= EX1valtmp;
          EX2val[STR2] <= EX2valtmp;
          EX3val[STR3] <= EX3valtmp;
          EX4val[STR4] <= EX4valtmp;
          EX5val[STR5] <= EX5valtmp;
          EX6val[STR6] <= EX6valtmp;
          EX7val[STR7] <= EX7valtmp;

          EX0cnt[STR0] <= EX0cnttmp;
          EX1cnt[STR1] <= EX1cnttmp;
          EX2cnt[STR2] <= EX2cnttmp;
          EX3cnt[STR3] <= EX3cnttmp;
          EX4cnt[STR4] <= EX4cnttmp;
          EX5cnt[STR5] <= EX5cnttmp;
          EX6cnt[STR6] <= EX6cnttmp;
          EX7cnt[STR7] <= EX7cnttmp;


          if ( oflo0 == 1'b0 ) SDL0in <= WG0in;    // No oflo, just transfer WGin to SDLin              
          else                                     // OFLO: apply clipping
            begin
            if ( sign0[STR0] == 1'b1 ) SDL0in <= 18'h1FFFF ; // use positive maximum
            else                       SDL0in <= 18'h20001 ; // use negative maximum          
            end              

          if ( oflo1 == 1'b0 ) SDL1in <= WG1in;    // No oflo, just transfer WGin to SDLin              
          else                                     // OFLO: apply clipping
            begin
            if ( sign1[STR1] == 1'b1 ) SDL1in <= 18'h1FFFF ; // use positive maximum
            else                       SDL1in <= 18'h20001 ; // use negative maximum          
            end              

          if ( oflo2 == 1'b0 ) SDL2in <= WG2in;    // No oflo, just transfer WGin to SDLin              
          else                                     // OFLO: apply clipping
            begin
            if ( sign2[STR2] == 1'b1 ) SDL2in <= 18'h1FFFF ; // use positive maximum
            else                       SDL2in <= 18'h20001 ; // use negative maximum          
            end              

          if ( oflo3 == 1'b0 ) SDL3in <= WG3in;    // No oflo, just transfer WGin to SDLin              
          else                                     // OFLO: apply clipping
            begin
            if ( sign3[STR3] == 1'b1 ) SDL3in <= 18'h1FFFF ; // use positive maximum
            else                       SDL3in <= 18'h20001 ; // use negative maximum          
            end              

          if ( oflo4 == 1'b0 ) SDL4in <= WG4in;    // No oflo, just transfer WGin to SDLin              
          else                                     // OFLO: apply clipping
            begin
            if ( sign4[STR4] == 1'b1 ) SDL4in <= 18'h1FFFF ; // use positive maximum
            else                       SDL4in <= 18'h20001 ; // use negative maximum          
            end              

          if ( oflo5 == 1'b0 ) SDL5in <= WG5in;    // No oflo, just transfer WGin to SDLin              
          else                                     // OFLO: apply clipping
            begin
            if ( sign5[STR5] == 1'b1 ) SDL5in <= 18'h1FFFF ; // use positive maximum
            else                       SDL5in <= 18'h20001 ; // use negative maximum          
            end              

          if ( oflo6 == 1'b0 ) SDL6in <= WG6in;    // No oflo, just transfer WGin to SDLin              
          else                                     // OFLO: apply clipping
            begin
            if ( sign6[STR6] == 1'b1 ) SDL6in <= 18'h1FFFF ; // use positive maximum
            else                       SDL6in <= 18'h20001 ; // use negative maximum          
            end              

          if ( oflo7 == 1'b0 ) SDL7in <= WG7in;    // No oflo, just transfer WGin to SDLin              
          else                                     // OFLO: apply clipping
            begin
            if ( sign7[STR7] == 1'b1 ) SDL7in <= 18'h1FFFF ; // use positive maximum
            else                       SDL7in <= 18'h20001 ; // use negative maximum          
            end              

          RAMWRT <= 1'b1;                     // set waveguide RAM write enable
          end

        4'hC:
          begin
          RAMWRT <= 1'b0;                     // reset waveguide RAM write enable

          SDL0adr[STR0] <= ( (WG0adr + 1) < MH0StrEnd ) ? (WG0adr + 1) : MH0StrStart ;
          SDL1adr[STR1] <= ( (WG1adr + 1) < MH1StrEnd ) ? (WG1adr + 1) : MH1StrStart ;
          SDL2adr[STR2] <= ( (WG2adr + 1) < MH2StrEnd ) ? (WG2adr + 1) : MH2StrStart ;
          SDL3adr[STR3] <= ( (WG3adr + 1) < MH3StrEnd ) ? (WG3adr + 1) : MH3StrStart ;
          SDL4adr[STR4] <= ( (WG4adr + 1) < MH4StrEnd ) ? (WG4adr + 1) : MH4StrStart ;
          SDL5adr[STR5] <= ( (WG5adr + 1) < MH5StrEnd ) ? (WG5adr + 1) : MH5StrStart ;
          SDL6adr[STR6] <= ( (WG6adr + 1) < MH6StrEnd ) ? (WG6adr + 1) : MH6StrStart ;
          SDL7adr[STR7] <= ( (WG7adr + 1) < MH7StrEnd ) ? (WG7adr + 1) : MH7StrStart ;

          MH0sum <= MH0sum + {{3{SDL0out[17]}},SDL0out};
          MH1sum <= MH1sum + {{3{SDL1out[17]}},SDL1out};
          MH2sum <= MH2sum + {{3{SDL2out[17]}},SDL2out};
          MH3sum <= MH3sum + {{3{SDL3out[17]}},SDL3out};
          MH4sum <= MH4sum + {{3{SDL4out[17]}},SDL4out};
          MH5sum <= MH5sum + {{3{SDL5out[17]}},SDL5out};
          MH6sum <= MH6sum + {{3{SDL6out[17]}},SDL6out};
          MH7sum <= MH7sum + {{3{SDL7out[17]}},SDL7out};

          if ( STR == 6 )
            begin
            if ( pass == 1 ) run <= 1'b0;        // on pass 1, stop the state machine after all strings are processed
            else 
              begin
              pass <= 1;                      // done with pass 0, set pass1
              state <= 4'h0;                  // restart at state 0
              end
            end
          else state <= 4'h1;
                    
          STR <= STR + 1;                     // select next string
          end

        endcase
        end
      end
    end  

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////  
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////  
// Decode structures for hardware receiving data from the MCU

// OUTPUTS FROM MIDI CONTROLLER
always @ ( posedge sysclk )
  begin
  if ( write_strobe )
    begin
    case ( port_id[1:0] )
    2'h0:
      begin
      Plucked_Str <= out_port[5:3];          // save the string's address for the appropriate state machine
      Plucked_MH  <= out_port[2:0];          // save the mini-harp's address for the appropriate state machine
      Set_Pluck <= 1'b1;                     // set the pluck occurred FF
      end
      
    2'h1: VEL          <= out_port[6:0];    
    2'h2: MOD_WHL      <= out_port[6:0];     // modulation wheel, global
    2'h3: SUSTAIN      <= out_port[6];       // SUSTAIN pedal
    endcase
    end
  else
    begin
    Set_Pluck <= 1'b0;                       // return set FF state to zero
    end
  end

// Instantiate the single Pluck_FF RS flip-flop
FDCPE #(.INIT(1'b0))
  Pluck_FF ( .Q( Pluck ),           // output
             .C( 1'b0 ),
             .CE( 1'b0 ),
             .CLR( Clr_Pluck ),     // FF clear signal
             .D( 1'b0 ),
             .PRE( Set_Pluck ) );   // FF set signal

// INPUTS TO MIDI CONTROLLER
  always @ ( posedge sysclk ) 
    begin
    case ( port_id[2:0] )                                    // decode and transfer data to in_port_reg
      3'h0: in_port_reg <= {1'b0,rx0ready_status,6'b000000}; // UART1 & UART0 rxready bits
      3'h1: in_port_reg <= rx0data;                          // MIDI UART rxdata
      3'h2: in_port_reg <= {4'b0000,SW[3:0]};                // slide switches, lower 4 used to select MIDI chan.
      3'h3: in_port_reg <= 8'd36;         // TRANSPOSE !!!  The MIDI controller will subtract this value 
                                          // from every note number received.  Use this to ensure that the 
                                          // lowest keyboard key plays the lowest string.

      default: in_port_reg <= 8'bxxxxxxxx;
    endcase
    end

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// LEDs

// DIAGNOSTIC:
//  assign led = port_id;    // hook LEDs to something to eliminate errors
//always @ ( posedge sysclk ) if ( EX0[0] == 1 ) DIAG <= DIAG + 1;

//assign led = EX0;
assign led = {2'b00,Plucked_Str,Plucked_MH};
//assign led = {PB,4'b0000};

//assign led = DIAG;
//always @ ( posedge sysclk )
//  begin
//  if ( Set_Pluck ) DIAG <= DIAG + 1;
//  end

endmodule
